/*
 *
 *  Copyright (C) 1998-99, OFFIS
 *
 *  This software and supporting documentation were developed by
 *
 *    Kuratorium OFFIS e.V.
 *    Healthcare Information and Communication Systems
 *    Escherweg 2
 *    D-26121 Oldenburg, Germany
 *
 *  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
 *  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
 *  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
 *  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
 *  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
 *
 *  Module: dcmpstat
 *
 *  Author: Marco Eichelberg
 *
 *  Purpose:
 *    classes: DVPSCurve_PList
 *
 *  Last Update:      $Author: meichel $
 *  Update Date:      $Date: 1998/12/22 17:57:05 $
 *  CVS/RCS Revision: $Revision: 1.1 $
 *  Status:           $State: Exp $
 *
 *  CVS/RCS Log at end of file
 *
 */

#ifndef __DVPSCUL_H__
#define __DVPSCUL_H__

#include "osconfig.h"    /* make sure OS specific configuration is included first */
#include "oflist.h"
#include "dctk.h"

class DVPSCurve;

/** the list of curves contained in image which is attached to a presentation state.
 *  This class manages the data structures comprising the list of curves
 *  (all instances of the Curve Module repeating elements)
 *  contained in an image object.
 */

class DVPSCurve_PList: private OFList<DVPSCurve *>
{
public:
  /// default constructor
  DVPSCurve_PList();
  
  /// copy constructor
  DVPSCurve_PList(const DVPSCurve_PList& copy);

  /** clone method.
   *  @return a pointer to a new DVPSCurve_PList object containing
   *  a deep copy of this object.
   */
  DVPSCurve_PList *clone() { return new DVPSCurve_PList(*this); }

  /// destructor
  virtual ~DVPSCurve_PList();

  /** reads all curve groups which can be displayed in a presentation state
   *  from a DICOM dataset.
   *  This method checks which curves are contained in the DICOM dataset.
   *  All curves complying with the presentation state requirements
   *  (i.e. 2-dimensional POLY or ROI curves without curve descriptor) are copied
   *  into the "list of curves" structure managed by this object.
   *  The completeness of the curves (presence of all required elements,
   *  value multiplicity) is checked.
   *  If this method returns an error code, the object is in undefined state afterwards.
   *  @param dset the DICOM dataset from which the curves are read
   *  @return EC_Normal if successful, an error code otherwise.
   */
  E_Condition read(DcmItem &dset);

  /** reset the object to initial state.
   *  After this call, the object is in the same state as after
   *  creation with the default constructor.
   */
  void clear();
  
  /** check presence of curve group
   *  @param group curve repeating group to be checked
   *  @return OFTrue if the specified curve group is present in the
   *     list of curves managed by this object.
   */
  OFBool haveCurveGroup(Uint16 group);

  /** get curve by group
   *  @param group curve repeating group to be checked
   *  @return a pointer to the matching DVPSCurve object if found,
   *    NULL otherwise.
   */
  DVPSCurve *getCurveGroup(Uint16 group);

  /** get curve by index
   *  @param idx index, must be < size()
   *  @return a pointer to the matching DVPSCurve object if it exists,
   *    NULL otherwise.
   */
  DVPSCurve *getCurve(size_t idx);

  /** get number of curves in this list.
   *  @return the number of curves.
   */
  size_t size() const { return OFList<DVPSCurve *>::size(); }  

};

#endif

/*
 *  $Log: dvpscul.h,v $
 *  Revision 1.1  1998/12/22 17:57:05  meichel
 *  Implemented Presentation State interface for overlays,
 *    VOI LUTs, VOI windows, curves. Added test program that
 *    allows to add curve data to DICOM images.
 *
 *
 */

